@@ -3,7 +3,7 @@ Connect to the `Anthropic API <https://console.anthropic.com/>`_ to use Claude s
 """
 
 # Copyright © 2024-2025 Frello Technology Private Limited
-# Copyright © 2025-2025 Yash Bonde github.com/yashbonde
+# Copyright © 2025- Yash Bonde github.com/yashbonde
 # MIT License
 
 from copy import deepcopy
@@ -209,7 +209,7 @@ class Anthropic(tt.ModelInterface):
         for line in lines_fn():
             if isinstance(line, bytes):
                 line = line.decode().strip()
-            if not line or not "data:" in line:
+            if not line or "data:" not in line:
                 continue
 
             resp = tu.from_json(line.replace("data:", "").strip())
@@ -253,6 +253,7 @@ class Anthropic(tt.ModelInterface):
                     input_tokens=usage_dict.pop("input_tokens"),
                     output_tokens=usage_dict.pop("output_tokens"),
                     cached_tokens=cached_tokens,
+                    model=self.model_id,
                     **usage_dict,
                 )
         if yield_usage:
@@ -320,6 +321,8 @@ class Anthropic(tt.ModelInterface):
         token: str | None = None,
         usage: bool = False,
         extra_headers: dict[str, str] | None = None,
+        debug: bool = False,
+        timeout=(5, 60),
         **kwargs,
     ):
         output = ""
@@ -331,8 +334,10 @@ class Anthropic(tt.ModelInterface):
             max_tokens=max_tokens,
             temperature=temperature,
             token=token,
+            debug=debug,
             usage=usage,
             extra_headers=extra_headers,
+            timeout=timeout,
             raw=False,
             **kwargs,
         ):
@@ -410,6 +415,8 @@ class Anthropic(tt.ModelInterface):
         token: str | None = None,
         usage: bool = False,
         extra_headers: dict[str, str] | None = None,
+        debug: bool = False,
+        timeout=(5, 60),
         **kwargs,
     ):
         output = ""
@@ -421,8 +428,10 @@ class Anthropic(tt.ModelInterface):
             max_tokens=max_tokens,
             temperature=temperature,
             token=token,
+            debug=debug,
             usage=usage,
             extra_headers=extra_headers,
+            timeout=timeout,
             raw=False,
             **kwargs,
         ):
@@ -564,6 +573,32 @@ class Anthropic(tt.ModelInterface):
         """This is the async function to convert speech to text"""
         raise NotImplementedError("Anthropic does not support speech to text")
 
+    def text_to_speech(
+        self,
+        prompt: str,
+        voice: str = "shimmer",
+        model="tts-1",
+        response_format="wav",
+        extra_headers: dict[str, str] | None = None,
+        timeout: tuple[int, int] = (5, 60),
+        **kwargs,
+    ) -> bytes:
+        """This is the blocking function to convert text to speech"""
+        raise NotImplementedError("Anthropic does not support text to speech")
+
+    async def text_to_speech_async(
+        self,
+        prompt: str,
+        voice: str = "shimmer",
+        model="tts-1",
+        response_format="wav",
+        extra_headers: dict[str, str] | None = None,
+        timeout: tuple[int, int] = (5, 60),
+        **kwargs,
+    ) -> bytes:
+        """This is the async function to convert text to speech"""
+        raise NotImplementedError("Anthropic does not support text to speech")
+
     # Batching methods
 
     def submit_batch(
