@@ -3,7 +3,7 @@ Connect to the `OpenAI API <https://playground.openai.com/>`_ and use their LLMs
 """
 
 # Copyright © 2024-2025 Frello Technology Private Limited
-# Copyright © 2025-2025 Yash Bonde github.com/yashbonde
+# Copyright © 2025- Yash Bonde github.com/yashbonde
 # MIT License
 
 import os
@@ -71,6 +71,10 @@ class OpenAIProtocol(tt.ModelInterface):
         final_messages = []
         for i, m in enumerate(thread.chats):
             if m.role == tt.Message.SYSTEM:
+                if i != 0:
+                    raise ValueError(
+                        "Only the first message in thread can be the system message."
+                    )
                 final_messages.append({"role": "system", "content": m.value})
             elif m.role == tt.Message.HUMAN:
                 if isinstance(m.value, str):
@@ -137,6 +141,24 @@ class OpenAIProtocol(tt.ModelInterface):
                 prev_tool_id = tu.get_random_string(5)  # reset tool id
             else:
                 raise Exception(f"Invalid message role: {m.role}")
+
+        if final_messages[0]["role"] == "system" and thread.tools:
+            tool_prompt = "# Tool Usage Instructions\n\n"
+            for tool in thread.tools:
+                if tool.system:
+                    tool_prompt += f"{tool.system}\n"
+            system = final_messages[0]["content"]
+            if isinstance(system, list):
+                system.append(
+                    {
+                        "type": "text",
+                        "text": tool_prompt,
+                    }
+                )
+            else:
+                system += tool_prompt
+            final_messages[0]["content"] = system
+
         return final_messages
 
     def _process_input(
@@ -171,21 +193,32 @@ class OpenAIProtocol(tt.ModelInterface):
         if extra_headers:
             headers.update(extra_headers)
         data = {
-            "temperature": temperature,
             "messages": final_messages,
             "model": model or self.model_id,
             "stream": stream,
         }
+        if not (
+            self.model_id.startswith("gpt-5")
+            or self.model_id.startswith("o4-")
+            or self.model_id.startswith("o3-")
+            or self.model_id.startswith("o1-")
+        ):
+            data["temperature"] = temperature
         if stream:
             data["stream_options"] = {"include_usage": usage}
         if max_tokens:
             data["max_tokens"] = max_tokens
-        if isinstance(chats, tt.Thread) and len(chats.tools):
+        if thread.tools:
             data["tools"] = [
                 {"type": "function", "function": x.to_dict()} for x in chats.tools
             ]
-            data["parallel_tool_calls"] = parallel_tool_calls
-        if isinstance(chats, tt.Thread) and chats.schema:
+            if not (
+                self.model_id.startswith("o4-")
+                or self.model_id.startswith("o3-")
+                or self.model_id.startswith("o1-")
+            ):
+                data["parallel_tool_calls"] = parallel_tool_calls
+        if thread.schema:
             resp_schema = chats.schema.model_json_schema()
             resp_schema["additionalProperties"] = False
             for _, defs in resp_schema.get("$defs", dict()).items():
@@ -245,6 +278,7 @@ class OpenAIProtocol(tt.ModelInterface):
                         input_tokens=usage.pop("prompt_tokens"),
                         output_tokens=usage.pop("completion_tokens"),
                         cached_tokens=usage["prompt_tokens_details"]["cached_tokens"],
+                        model=self.model_id,
                         **usage,
                     )
 
@@ -328,10 +362,19 @@ class OpenAIProtocol(tt.ModelInterface):
                 raw=False,
                 debug=debug,
                 timeout=timeout,
+                usage=usage,
                 **kwargs,
             ):
                 if isinstance(x, dict):
-                    output = x
+                    if "name" in x and "arguments" in x:
+                        tool_fn = list(
+                            filter(lambda tool: tool.name == x["name"], chats.tools)
+                        )
+                        if not tool_fn:
+                            raise ValueError(f"Tool {x['name']} not found in thread")
+                        output = tt.ToolCall(tool=tool_fn[0], arguments=x["arguments"])
+                    else:
+                        output = x
                 elif isinstance(x, tt.Usage):
                     usage_obj = x
                 else:
@@ -340,7 +383,7 @@ class OpenAIProtocol(tt.ModelInterface):
             tu.logger.error(f"API Error: {e.response.text}")
             raise e
 
-        if isinstance(chats, tt.Thread) and chats.schema:
+        if isinstance(chats, tt.Thread) and chats.schema and isinstance(output, str):
             output = chats.schema(**tu.from_json(output))
 
         if usage:
@@ -447,10 +490,19 @@ class OpenAIProtocol(tt.ModelInterface):
                 extra_headers=extra_headers,
                 raw=False,
                 timeout=timeout,
+                usage=usage,
                 **kwargs,
             ):
                 if isinstance(x, dict):
-                    output = x
+                    if "name" in x and "arguments" in x:
+                        tool_fn = list(
+                            filter(lambda tool: tool.name == x["name"], chats.tools)
+                        )
+                        if not tool_fn:
+                            raise ValueError(f"Tool {x['name']} not found in thread")
+                        output = tt.ToolCall(tool=tool_fn[0], arguments=x["arguments"])
+                    else:
+                        output = x
                 elif isinstance(x, tt.Usage):
                     usage_obj = x
                 else:
@@ -459,7 +511,7 @@ class OpenAIProtocol(tt.ModelInterface):
             tu.logger.error(f"API Error: {e.response.text}")
             raise e
 
-        if isinstance(chats, tt.Thread) and chats.schema:
+        if isinstance(chats, tt.Thread) and chats.schema and isinstance(output, str):
             output = chats.schema(**tu.from_json(output))
 
         if usage:
