@@ -12,6 +12,9 @@ from tuneapi.types.chats import (
     ThreadsTree,
     Prop,
     Tool,
+    tool,
+    AsyncInSyncError,
+    ToolCall,
     EmbeddingGen,
     WebVTTCue,
     Transcript,
@@ -27,4 +30,9 @@ from tuneapi.types.evals import (
     Evals,
 )
 
-from tuneapi.types.bm import BM, F, Field
+from tuneapi.types.bm import (
+    BM,
+    F,
+    Field,
+    Bool,
+)
